/** SigusEffect.java:
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
 @author Hemerson Pistori (pistori@ec.ucdb.br)
 @created 18 de Maro de 2004
 
*/

package sigus.core;

import sigus.machineLearning.*;
import sigus.parameterExtraction.*;


import javax.media.*;
import javax.media.format.*;
import java.awt.*;
import java.util.*;


import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import ij.io.*;

/**
 *  A SigusEffect integrates all the modules of a visual based system, which include some digital image
 *  processing filters, a feature extractor and a machine learning module.
*/
abstract public class SigusEffect implements Effect {

    /** Format of the input image (e.g. 8-bit, RGB, HSV, etc)*/
    Format    inputFormat;
    /** Format of the output image (e.g. 8-bit, RGB, HSV, etc)*/
    Format    outputFormat;
    /** Supported input formats (the formats that the effect can process) */
    Format[]  inputFormats;
    /** Support output formats (the formats that the effect can generate)*/
    Format[]  outputFormats;

    /** Application that will process the recognized visual signs */
    public SigusApp sigusApp = null;

    /** Algorithm that will be used to learn from examples */
    public Learner_ learner = null;

    /** The updater reads each training sample and writes it, converts it to arff (weka) format, and writes it to an arff file */
    public Updater updater = null;

    /** Extracts the parameters from a segmented and ROI marked image */
    public ParameterExtractor_ pe = null;

    /** The number of different classes + 1 (e.g.: TTT has 9 classes - the 9          different screen positions. The LIBRAS Prototype has  36 classes: all 26         lower case letters plus 10 digits. The aditional class is used to indicate       "no class detected yet" */
    public int MAX_CLASS;

    /** The last identified class (e.g. the last "signed" letter or the last square where the eyes looked at). */
    int lastFocus = MAX_CLASS;

    /** The current identified class (e.g. the "signed" letter or the last square where the eyes looked at). */
    int focus = MAX_CLASS;

    /** Counter for the number of times that the same sign has been recognizedThe current identified class (e.g. the "signed" letter or the last square where the eyes looked at). */
    int sameFocusCounter = 0;

    /** In order to be recognized, a sign must be detect in "maxFocus" consecutive frames. */
    public int maxFocus = 1;

    /** The number of training instances that should be collected before the learning algorithm is called for the first time */
    public int minInstances = 2000;

    /** True if the frames captured (before and after processing) should be saved on disk */
    public boolean saveInstancesImages = false;

    /** True if system is in the calibration phase (e.g. extracting samples from human skin */
    boolean calibrating = true;

    /** ImageJ image processor. Unfortunately, we didn't find a way to work directly in the JMF Image Window using ImageJ tools. That is why a second Image Window is created and pointed by ip and imp. It is in this second window most part of the image processing happens */
    public static ImageProcessor ip = null;

    /** ImageJ window where the processed image will be presented */
    public static ImagePlus imp = null;


    int[] imageCopy = null; // Used when a copy of the image must be saved. This vector was created in order to use the available ImageJ saving methods.

    /** Auxiliary ImageProcessor used when images must be saved on disk  */
    static ImageProcessor ipCopy = null;

    /** Auxiliary ImagePlus used when images must be saved on disk  */
    static ImagePlus impCopy = null;

    /** Coordinates of the last ROI - Region of Intereset */
    Rectangle lastRoi = null;


    /** Set focus related variables and supported formats */
    public SigusEffect() {
        init();
        lastFocus = MAX_CLASS;
        focus = MAX_CLASS;
        sameFocusCounter = 0;
        inputFormats = new Format[]{
                           new RGBFormat( null,
                                          Format.NOT_SPECIFIED,
                                          //Format.intArray,
                                          Format.byteArray,
                                          Format.NOT_SPECIFIED,
                                          24,
                                          3, 2, 1,
                                          3, Format.NOT_SPECIFIED,
                                          Format.TRUE,
                                          Format.NOT_SPECIFIED )
                       };

        outputFormats = new Format[]{
                            new RGBFormat( null,
                                           Format.NOT_SPECIFIED,
                                           //Format.intArray,
                                           Format.byteArray,
                                           Format.NOT_SPECIFIED,
                                           24,
                                           3, 2, 1,
                                           3, Format.NOT_SPECIFIED,
                                           Format.TRUE,
                                           Format.NOT_SPECIFIED )
                        };

    }


    /**
     Takes a input image, inBuffer, and produces a processed image,outBuffer. Control the learning process.

     @param  inBuffer   Buffer containing image captured by CameraReader.class
     @param  outBuffer  Buffer that will be returned to CameraReader.class
     @return            Status of the process (ok or not ok).
     */
    public int process( Buffer inBuffer, Buffer outBuffer ) {

        int outputDataLength  = ( (VideoFormat) outputFormat ).getMaxDataLength();
        validateByteArraySize( outBuffer, outputDataLength );

        outBuffer.setLength( outputDataLength );
        outBuffer.setFormat( outputFormat );
        outBuffer.setFlags( inBuffer.getFlags() );

        byte[]     inData            = (byte[]) inBuffer.getData();
        byte[]     jmfData           = (byte[]) outBuffer.getData();
        byte[]     outData           = null;
        RGBFormat  vfIn              = (RGBFormat) inBuffer.getFormat();
        Dimension  sizeIn            = vfIn.getSize();

        int        width             = (int) sizeIn.getWidth(); // Image Width
        int        height            = (int) sizeIn.getHeight(); // Image Height
        int        offset = width*3;  // Three bytes are used to represented a pixel in the RGB format. The offset is the
        // size of a row in the image byte array. It is used to access the buffer as a two
        // dimensional array.

        imageCopy = null; // Used when a copy of the image must be saved. This vector was created in order to use the available ImageJ saving methods.

        byte value;


        if(sigusApp != null)
            calibrating = sigusApp.getCalibrating();

        if( saveInstancesImages )
            imageCopy = new int[width*height];  // Used just to save a copy of the input image. This vector was created in order to use the
        // available ImageJ saving methods.

        createProcessedImageWindow(width,height);


        imp.killRoi(); // Reset the ImageJ ROI (A rectangular region in the image)


        if(ip != null) {

            outData = (byte[]) ip.getPixels();

            if( saveInstancesImages )
                copyImage(inData,imageCopy,width,height);


            applyFilters(inData,jmfData,outData,width,height);
            imp.updateAndDraw();

            if( saveInstancesImages )
                saveImage(inData,outData,width,height);


            if(pe!=null) {
                pe.extract(imp);
                if(updater!=null && updater.isUpdating()) {
                    if(!pe.isBadSample()) {
                        updater.update();
                    }
                }
            }

        } else {
            System.out.println("Problems creating the ImageJ Processed Image Window");
        }

        // Initialize an Updater and a ParameterExtractor
        if(updater == null) {
            createParameterExtractor();
            IJ.showStatus("Loading Updater...");
            createUpdater();
            updater.setImageProcessor(ip);
            updater.setImagePlus(imp);
            updater.setParameterExtractor(pe);
            IJ.showStatus("UpdateTrainingSet Loading Finished");
        }

        // Initialize a Learner
        if(learner == null && (updater.getTotalInstances() == minInstances || updater.getStopRequested())) {
            IJ.showStatus("Loading Learner...");
            sigusApp.setGeneralTitle("Wait... Learning from Instances...");
            learner = new Learner_();
            if (learner != null) {
                learner.init(updater.getFileName());
                updater.stopUpdating();
                updater.setLearner(learner);
                sigusApp.setTitlePlaying();
                //				updater.stopUpdating();
            }
            IJ.showStatus("Learner Loading Finished");
        }

        // Initialize the Application
        if(sigusApp == null) {
            IJ.showStatus("Loading Application...");
            createSigusApp();
            sigusApp.setUpdater(updater);
            sigusApp.setImageProcessor(ip);
            sigusApp.init();
            IJ.showStatus("Application Loading Finished");
        }

        if( ip != null ) {
            if(learner != null && learner.isReady() && updater!=null && !updater.isUpdating()) {
                lastFocus = focus;
                focus = learner.getFocus(imp,pe);
                if(focus == 0)
                    sameFocusCounter = 0;
            } else {
                sameFocusCounter = 0;
            }
        } else {
            System.out.println("ERROR: ImageJ viewer is not set");
            sameFocusCounter = 0;
        }


        if(lastFocus == MAX_CLASS) {
            lastFocus = focus;
            sameFocusCounter = 0;
        } else if(lastFocus != focus) {
            sameFocusCounter = 0;
        } else {
            ++sameFocusCounter;
            if(sameFocusCounter > maxFocus && focus > 0 && focus < MAX_CLASS) {
                sigusApp.process(focus);
                sameFocusCounter = 0;
            }
        }

        return BUFFER_PROCESSED_OK;
    }


    /** Start or stop saving images to disk  */
    public void setSaveInstances(boolean value) {
        saveInstancesImages = value;
    }


    /** Create the window where the processed image is shown */
    void createProcessedImageWindow(int width, int height) {
        if(imp == null) {
            ip = new ByteProcessor(width, height);
            imp = (ImagePlus) new ImagePlus("Processed Image", ip);
            imp.show();
            imp.getWindow().setLocation(0,height+60);
        }
    }

    /** Copy the image to the imageCopy vector
     @param  inBuffer   Buffer containing image captured by CameraReader.class
     @param  outBuffer  Buffer containing the processed image
     @param  width      Image width
     @param  height     Image height
    */
    void copyImage(byte[] inData, int[] imageCopy, int height, int width) {

        int  offset = width*3;  // Three bytes are used to represented a pixel in the RGB format. The offset is the
        // size of a row in the image byte array. It is used to access the buffer as a two
        // dimensional array.
        int R,G,B;


        for(int y = 0; y < height ; y++ ) {
            for( int x = 0; x < width; x++ ) {
                int k  = x*3;

                R = inData[k+2+y*offset]&0xff;
                G = inData[k+1+y*offset]&0xff;
                B = inData[k+y*offset]&0xff;

                // The following line is converting from JMF format (sequence of byte triples - R, G, B ) to
                // ImageJ ColorProcessor format (sequence of integers - each int encoding the tree primitive colors).
                // The image must also be vertically flipped and the sequence of RGB primitives are
                // also different from JMF to ImageJ.
                imageCopy[x+((height-1)-y)*width] = 0xff000000 | (((byte)R&0xff)<<16) | (((byte)G&0xff)<<8) | (byte)B&0xff;
            }
        }

    }




    /** Save two images to disk: before and after processing

     @param  inBuffer   Buffer containing image captured by CameraReader.class
     @param  outBuffer  Buffer containing the processed image
     @param  width      Image width
     @param  height     Image height
    */
    void saveImage(byte[] inData, byte[] outData,int width,int height) {


        System.out.println("Saving image "+updater.getTotalInstances()+" in the Parent Directory");

        if(impCopy == null) {

            ipCopy = new ColorProcessor(width, height);
            impCopy = (ImagePlus) new ImagePlus("Input Image", ipCopy);

        }

        ipCopy.setPixels(imageCopy);

        new FileSaver(impCopy).saveAsTiff("exemploAntes"+
                                          updater.getClassValue()+
                                          "-" +
                                          updater.getTotalInstances()+
                                          ".tif");

        new FileSaver(imp).saveAsTiff("exemploDepois"+
                                      updater.getClassValue()+
                                      "-" +
                                      updater.getTotalInstances()+
                                      ".tif"
                                     );

    }


    public boolean isCalibrating() {
        return calibrating;
    }


    public void resetLearner() {
        if(learner != null) {
            learner = null;
        }
        if(updater != null) {
            updater.setLearner(null);
            updater.clearTrainingSet();
            updater.setStopRequested(false);
        }
    }

    /** The Digital Image Processing algorithms should be implemented or started in this function */
    abstract public void applyFilters(byte[] inData, byte[] jmfData, byte[] outData, int width, int height);
    abstract public void createParameterExtractor();
    abstract public void createUpdater();
    abstract public void createSigusApp();
    abstract public void init();

    public String getName() {
        return "Effect Name";
    }

    public void open() {}
    public void close() {}
    public void reset() {}
    public Object getControl( String controlType ) {
        return null;
    }
    public Object[] getControls() {
        return null;
    }

    public Format[] getSupportedInputFormats() {
        return inputFormats;
    }

    /**
     Get the supported output formats. 

     @param  input      If input is not null, return only the output formats that are compatible to the input format.
     @return            Supported formats 
     */
    public Format[] getSupportedOutputFormats( Format input ) {
        if ( input == null ) {
            return outputFormats;
        }

        if ( matches( input, inputFormats ) != null ) {
            return new Format[]{outputFormats[0].intersects( input )};
        } else {
            return new Format[0];
        }
    }


    public Format setInputFormat( Format input ) {
        inputFormat = input;
        return input;
    }


    public Format setOutputFormat( Format output ) {
        if ( output == null || matches( output, outputFormats ) == null ) {
            return null;
        }
        RGBFormat  incoming       = (RGBFormat) output;

        Dimension  size           = incoming.getSize();
        int        maxDataLength  = incoming.getMaxDataLength();
        int        lineStride     = incoming.getLineStride();
        float      frameRate      = incoming.getFrameRate();
        int        flipped        = incoming.getFlipped();
        int        endian         = incoming.getEndian();

        if ( size == null ) {
            return null;
        }
        if ( maxDataLength < size.width * size.height * 3 ) {
            maxDataLength = size.width * size.height * 3;
        }
        if ( lineStride < size.width * 3 ) {
            lineStride = size.width * 3;
        }
        if ( flipped != Format.FALSE ) {
            flipped = Format.FALSE;
        }

        outputFormat = outputFormats[0].intersects( new RGBFormat( size,
                       maxDataLength,
                       null,
                       frameRate,
                       Format.NOT_SPECIFIED,
                       Format.NOT_SPECIFIED,
                       Format.NOT_SPECIFIED,
                       Format.NOT_SPECIFIED,
                       Format.NOT_SPECIFIED,
                       lineStride,
                       Format.NOT_SPECIFIED,
                       Format.NOT_SPECIFIED ) );

        //System.out.println("final outputformat = " + outputFormat);
        return outputFormat;
    }



    /**
     * Test if a specific format is compatible to any format from the list of supported formats
     *
     * @param  in    format to be tested
     * @param  outs  list of supported formats
     * @return       supported format compatible to in
     */
    Format matches( Format in, Format outs[] ) {
        for ( int i = 0; i < outs.length; i++ ) {
            if ( in.matches( outs[i] ) ) {
                return outs[i];
            }
        }

        return null;
    }


    /**
     *  Verify if buffer can be properly converted to a byte array.
     *
     * @param  buffer   buffer containing the image.
     * @param  newSize  the size of the byte array
     * @return          image converted to a byte array.
     */
    byte[] validateByteArraySize( Buffer buffer, int newSize ) {
        Object  objectArray  = buffer.getData();
        byte[]  typedArray;

        if ( objectArray instanceof byte[] ) {     // is correct type AND not null
            typedArray = (byte[]) objectArray;
            if ( typedArray.length >= newSize ) {     // is sufficient capacity
                return typedArray;
            }

            byte[]  tempArray  = new byte[newSize];     // re-alloc array
            System.arraycopy( typedArray, 0, tempArray, 0, typedArray.length );
            typedArray = tempArray;
        } else {
            typedArray = new byte[newSize];
        }

        buffer.setData( typedArray );
        return typedArray;
    }

}

