/** ColorBasedSegmentation.java:
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
 @author Hemerson Pistori (pistori@ec.ucdb.br)
 @created 18 de Maro de 2004
 
*/

package sigus.segmentation;

import sigus.*;


import javax.media.*;
import javax.media.format.*;
import java.awt.*;
import java.util.*;


import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import ij.io.*;

/**
 
Implements a color based segmentation technique based on statistichal models of the target object color. 
*/
public class ColorBasedSegmentation {

    /** Dimension of the color space. Initialized in 2 because the model will be based on two variables: Hue and Saturation */
    int dimension = 2;

    /** Dimensions and position of the rectangular window, in the training image, where the samples will be collected */
    Rectangle targetSampleWindow = new Rectangle(150,160,25,20);

    /** Mean values for each variable*/
    double[] muValues = new double[dimension];
    /** Sigma values for each variable */
    double[] sigmaValues = new double[dimension];
    /** Maximum Number of samples collected. It is assumed that the same number of samples will be collected for each variable */
    int N = targetSampleWindow.getSize().width*targetSampleWindow.getSize().height;
    /** Samples collected so far */
    int nn = 0;

    /** Variable Values */
    int[][] values = new int[dimension][N];

    /** The constant that should be multiplied to the standard deviation when calculating the interval, for each variable, that will be used to identify target pixels */
    double stdMult = 1.8;

    /** True when mean e standard deviation parameters have been estimated */
    private boolean estimated = false;

    public ColorBasedSegmentation() { }


    /**
       Takes an input image, inData, and creates and output image, outData, with all pixels that are not part of the target object (a hand, a mice, etc), set to zero 

       @param  inBuffer   Input Image 
       @param  outBuffer  output Image
       @param  width      Image width
       @param  height     Image height
     */

    public void segment(byte[] inData, byte[] jmfData, byte[] outData, int width, int height) {

        int        offset = width*3;
        int        R,G,B;

        if(!estimated) {
            estimateParameters();
            estimated = true;
        }


        for(int y = 0; y < height ; y++ ) {
            for( int x = 0; x < width; x++ ) {
                int k  = x*3;

                R = inData[k+2+y*offset]&0xff;
                G = inData[k+1+y*offset]&0xff;
                B = inData[k+y*offset]&0xff;


                if( isTargetColor(R,G,B) ) {
				
			
                    outData[x + ((height-1)-y)*width] = (byte)
                                                        ((int)(R*0.299 + G*0.587 + B*0.114)&0xff);
														


                    jmfData[k+y*offset] = inData[k+y*offset];
                    jmfData[k+1+y*offset] = inData[k+1+y*offset];
                    jmfData[k+2+y*offset] = inData[k+2+y*offset];

                } else {

                    outData[x + ((height-1)-y)*width] = 0;

                    jmfData[k+y*offset] = 0;
                    jmfData[k+1+y*offset] = 0;
                    jmfData[k+2+y*offset] = 0;

                }
            }

        }

    }

    /**
        Calculate Means and Standard Deviations values for all variables
    */
    void estimateParameters() {


        for(int i = 0; i < dimension; ++i ) {
            muValues[i] = muValues[i]/N;
        }

        for(int i = 0; i < dimension; ++i ) {
            for(int j = 0; j < N; ++j ) {
                sigmaValues[i] = sigmaValues[i] + Math.pow((values[i][j] - muValues[i]),2);

            }
        }

        for(int i = 0; i < dimension; ++i ) {
            sigmaValues[i] = Math.sqrt(sigmaValues[i]/(N-1));
        }

    }


    /**
        True if H and S value of the pixel are inside the calculated limits
    */
    boolean isTargetColor(int R, int G, int B) {

        float[] hsb = new float[3];


           hsb = Color.RGBtoHSB(R, G, B, hsb);
		   

           boolean isTarget = true;


        for(int i = 0; i < dimension; ++i ) {
            double lowerLimit = muValues[i]-(stdMult*sigmaValues[i]);
            double upperLimit = muValues[i]+(stdMult*sigmaValues[i]);

            if( ((hsb[i]*255.0) < lowerLimit) || ((hsb[i]*255.0) > upperLimit))
                isTarget = false;

        }

        return isTarget;

    }



    /**
        Collect H and S values 
    */
    public void extractSamples(byte[] inData, byte[] jmfData, byte[] outData, int width, int height) {

        int offset = width*3;


        float[] hsb = new float[3];
        int R,G,B;

        clearSamplesAndParameters();


        for(int y = 0; y < height ; y++ ) {
            for( int x = 0; x < width; x++ ) {
                int k  = x*3;

                R = inData[k+2+y*offset]&0xff;
                G = inData[k+1+y*offset]&0xff;
                B = inData[k+y*offset]&0xff;

                if(targetSampleWindow.contains(x,y)) {
                    hsb = Color.RGBtoHSB(R, G, B, hsb);

                    collectSample(hsb);


                    jmfData[k+y*offset] = (byte) (inData[k+y*offset]-30); // sample window will be a little lighter (+ 50)
                    jmfData[k+1+y*offset] = (byte) (inData[k+1+y*offset]-30); // sample window will be a little lighter (+ 50)
                    jmfData[k+2+y*offset] = (byte) (inData[k+2+y*offset]-30); // sample window will be a little lighter (+ 50)

                    // Converts to gray scale
                    outData[x + ((height-1)-y)*width] = (byte) (((int)(R*0.299+G*0.587+B*0.114)&0xff)+20);


                    outData[x + ((height-1)-y)*width] = (byte) x;


                } else {

                    jmfData[k+y*offset] = inData[k+y*offset];
                    jmfData[k+1+y*offset] = inData[k+1+y*offset];
                    jmfData[k+2+y*offset] = inData[k+2+y*offset];

                    // Converts to gray scale
                    outData[x + ((height-1)-y)*width] = (byte) ((int)(R*0.299+G*0.587+B*0.114)&0xff) ;

                    //                    outData[x + ((height-1)-y)*width] = (byte) G;


                }
            }
        }
    }


    void clearSamplesAndParameters() {

        nn = 0;
        estimated = false;

        for(int i = 0; i < dimension; ++i ) {
            muValues[i] = 0;
            sigmaValues[i] = 0;

            for(int j = 0 ; j < N; ++j ) {
                values[i][j] = 0;
            }

        }

    }


    void collectSample( float[] hsb) {


        for(int i = 0; i < dimension; ++i ) {
            values[i][nn] = (int) (hsb[i]*255.0);
            muValues[i] += hsb[i]*255.0;
        }

        ++nn;
    }


}

