import ij.*;
import ij.plugin.filter.PlugInFilter;
import ij.process.*;
import java.awt.*;
import ij.gui.*;


/** HoughLines_.java
 
This ImageJ plugin shows the Hough Transform Space for lines
of predefined radius. The source image must be an 8-Bit black & white.
<p>
Houghlines_.java is open-source. You are free to do anything you want
with this source as long as I get credit for my work.
 
@author Hemerson Pistori (pistori@ec.ucdb.br)
      e Eduardo Rocha Costa (eduardo.rocha@poli.usp.br)
 
*/

public class HoughLines_ implements PlugInFilter {

    byte imageValues[]; // Raw image (returned by ip.getPixels())
    int houghValues[][]; // Hough Space Values
    public int width; // Original Image width
    public int height;  // Original Image height
    public int hWidth; // HoughSpace Width
    public int hHeight;  // HoughSpace Height
    public int offset; // Image Width
    public int offx;   // ROI x offset
    public int offy;   // ROI y offset
    private int maxLines = 10;
    private boolean useThreshold = true;
    private int threshold = 40; // Percent related to the maximum hough value.
    private int centre_x;
    private int centre_y;
    private double theta_step;
    private boolean showProgressBar = true;
    Point bestLines[]; // Parameters the lines Found.


    public int setup(String arg, ImagePlus imp) {
        if (arg.equals("about")) {
            showAbout();
            return DONE;
        }
        return DOES_8G+DOES_STACKS+SUPPORTS_MASKING;
    }

    public void run(ImageProcessor ip) {


        if( readParameters() ) { // Show a Dialog Window for user input of
            // radius and maxCircles.


            imageValues = (byte[])ip.getPixels();
            Rectangle r = ip.getRoi();

            offx = r.x;
            offy = r.y;
            width = r.width;
            height = r.height;
            offset = ip.getWidth();

            houghTransform();

            // Create image View for Hough Transform.
            ImageProcessor newip = new ByteProcessor(hWidth, 2*hHeight);
            byte[] newpixels = (byte[])newip.getPixels();
            createHoughPixels(newpixels);


            // Create image View for Marked Lines
            ImageProcessor linesip = new ByteProcessor(width, height);
            byte[] linesPixels = (byte[])linesip.getPixels();

            // Draw the lines in a new image
            // getBestLines(maxLines);
            drawLines(linesPixels);

            new ImagePlus("Hough Space for Lines", newip).show();
            new ImagePlus("Lines Found", linesip).show();
        }

    }


    boolean readParameters() {

        GenericDialog gd = new GenericDialog("Hough Parameters", IJ.getInstance());
        gd.addNumericField("Threshold: (Higher Values, More Lines)", 40, 0);
        gd.showDialog();

        if (gd.wasCanceled()) {
            return(false);
        }

        //			maxLines = (int) gd.getNextNumber();
        threshold = (int) gd.getNextNumber();
        /*			if (maxLines > 0) {
        				useThreshold = false;
        				threshold = -1;
        		    } else {
        				useThreshold = true;
        				if(threshold < 0) {
        	               IJ.showMessage("Threshold must be greater than 0");
        	               return(false);
        			    }
        		    }*/
        return(true);

    }


    void showAbout() {
        IJ.showMessage("About lines_...",
                       "This plugin finds straight lines\n" +
                       "using a basic HoughTransform operator\n." +
                       "For better results apply an Edge Detector\n" +
                       "a Binarizer (Threshold) and Invert the Image" +
                       "before using this plugin\n"+
                       "\nAuthor: Hemerson Pistori (pistori@ec.ucdb.br)"
                      );
    }


    // The hard work is done here.
    private void houghTransform () {


        hWidth = 500;
        centre_x = width / 2;
        centre_y = height / 2;
        byte pixelValue;

        theta_step = Math.PI / hWidth;
        int tmp = Math.max(height,width);
        hHeight = (int) (Math.sqrt(2)*tmp);

        //Create the hough array and initialize to zero
        houghValues = new int [hWidth][2*hHeight];
        for(int i = 0; i < hWidth; i++) {
            for(int j = 0; j < 2*hHeight; j++) {
                houghValues[i][j] = 0;
            }
        }


        //Update hough array
        for(int i = 0; i < width; i++) {

            showProgress((double)i/width);

            for(int j = 0; j < height; j++) {
                //	         pixelValue = imageValues[((height-j)-1)*width + i];
                pixelValue = imageValues[j*width + i];
                if(pixelValue != 0) {
                    for(int k = 0; k < hWidth; k++) {

                        //Work out the r values for each theta step
                        tmp = (int) (((i-centre_x)*Math.cos(k*theta_step)) +
                                     ((j-centre_y)*Math.sin(k*theta_step)));

                        tmp = tmp + hHeight;
                        if (tmp < 0 || tmp >= 2*hHeight)
                            continue;

                        //Increment hough array
                        houghValues[k][tmp]++;
                    }
                }
            }
        }
    }




    // Convert Values in Hough Space to an 8-Bit Image Space.
    private void createHoughPixels (byte houghPixels[]) {
        //Convert the output array from 2_d to 1_d
        for(int i = 0; i < hWidth; i++) {
            for(int j = 0; j < 2*hHeight; j++) {
                houghPixels[j*hWidth + i] = (byte) houghValues[i][j];
            }
        }
    }


    // Draw the lines found in the original image.
    public void drawLines(byte[] linesPixels) {

        int high = 0;

        int thresh;

        //Find the max hough value for the thresholding operation
        for(int i = 0; i < hWidth; i++) {
            for(int j = 0; j < 2*hHeight; j++) {
                if(houghValues[i][j] > high) {
                    high = houghValues[i][j];
                }
            }
        }

        //Set the threshold limit
        thresh = high*(100-threshold)/100;

        // Search for local peaks above threshold to draw
        boolean draw = false;
        int k;
        int l;
        int dt;     // test theta
        int dr;     // test offset
        for(int i = 0; i < hWidth; i++) {
            for(int j = 0; j < 2*hHeight; j++) {

                // only consider points above threshold
                if(houghValues[i][j] >= thresh) {

                    // see if local maxima
                    draw = true;
                    int peak = houghValues[i][j];
                    for(k = -1; k < 2; k++) {
                        for(l = -1; l < 2; l++) {
                            if (k==0 && l==0)
                                continue;
                            dt = i+k;
                            dr = j+l;
                            if (dr < 0 || dr >= 2*hHeight)
                                continue;
                            if (dt < 0)
                                dt = dt + hWidth;
                            if (dt >= hWidth)
                                dt = dt - hWidth;
                            if (houghValues[dt][dr] > peak) {
                                draw = false;
                                break;
                            }
                        }
                    }
                    if (!draw)
                        continue;



                    //Draw edges in output array
                    double tsin = Math.sin(i*theta_step);
                    double tcos = Math.cos(i*theta_step);
                    if (i <= hWidth/4 || i >= (3*hWidth)/4) {
                        for(int y = 0; y < height; y++) {
                            int x = (int) (((j-hHeight) - ((y-centre_y)*tsin)) / tcos) + centre_x;
                            if(x < width && x >= 0) {
                                linesPixels[y*width+x] = -1;
                            }
                        }
                    } else {
                        for(int x = 0; x < width; x++) {
                            int y = (int) (((j-hHeight) - ((x-centre_x)*tcos)) / tsin) + centre_y;
                            if(y < height && y >= 0) {
                                linesPixels[y*width+x] = -1;
                            }
                        }
                    }
                }
            }
        }
    }


    private void showProgress(double progress) {
        if (showProgressBar)
            IJ.showProgress(progress);
    }

    /*
    	public Point nthMaxLine(int i)
    	{
    		return bestLines[i];
    	}
     
     
    	private void getBestLines (int i)
    	{
        }*/
}

